@@ -1,4 +1,5 @@
 """Module for handling the Amazon service"""
+
 import base64
 import hashlib
 import json
@@ -7,13 +8,21 @@ import os
 import secrets
 import struct
 import time
+import urllib
 import uuid
+from collections import defaultdict
 from gettext import gettext as _
+from typing import Any, Dict, Optional
 from urllib.parse import parse_qs, urlencode, urlparse
 
+import yaml
+
 from lutris import settings
-from lutris.exceptions import AuthenticationError, UnavailableGame
-from lutris.services.base import OnlineService
+from lutris.exceptions import AuthenticationError, UnavailableGameError
+from lutris.installer import AUTO_WIN32_EXE
+from lutris.installer.installer_file import InstallerFile
+from lutris.installer.installer_file_collection import InstallerFileCollection
+from lutris.services.base import SERVICE_LOGIN, OnlineService
 from lutris.services.service_game import ServiceGame
 from lutris.services.service_media import ServiceMedia
 from lutris.util import system
@@ -25,19 +34,21 @@ from lutris.util.strings import slugify
 
 class AmazonBanner(ServiceMedia):
     """Game logo"""
+
     service = "amazon"
     size = (200, 112)
     dest_path = os.path.join(settings.CACHE_DIR, "amazon/banners")
-    file_pattern = "%s.jpg"
+    file_patterns = ["%s.jpg"]
     api_field = "image"
     url_pattern = "%s"
 
-    def get_media_url(self, details):
+    def get_media_url(self, details: Dict[str, Any]) -> Optional[str]:
         return details["product"]["productDetail"]["details"]["logoUrl"]
 
 
 class AmazonGame(ServiceGame):
     """Representation of a Amazon game"""
+
     service = "amazon"
 
     @classmethod
@@ -55,24 +66,25 @@ class AmazonService(OnlineService):
     """Service class for Amazon"""
 
     id = "amazon"
-    name = _("Amazon Prime Gaming")
+    name = _("Amazon")
     icon = "amazon"
+    runner = "wine"
     has_extras = False
     drm_free = False
-    medias = {
-        "banner": AmazonBanner
-    }
+    medias = {"banner": AmazonBanner}
     default_format = "banner"
 
     login_window_width = 400
     login_window_height = 710
 
     marketplace_id = "ATVPDKIKX0DER"
-    user_agent = "com.amazon.agslauncher.win/2.1.7437.6"
+    user_agent = "com.amazon.agslauncher.win/3.0.9202.1"
 
     amazon_api = "https://api.amazon.com"
     amazon_sds = "https://sds.amazon.com"
     amazon_gaming_graphql = "https://gaming.amazon.com/graphql"
+    amazon_gaming_distribution = "https://gaming.amazon.com/api/distribution/v2/public"
+    amazon_gaming_entitlements = "https://gaming.amazon.com/api/distribution/entitlements"
 
     client_id = None
     serial = None
@@ -86,8 +98,6 @@ class AmazonService(OnlineService):
 
     locale = "en-US"
 
-    is_loading = False
-
     @property
     def credential_files(self):
         return [self.user_path, self.cookies_path]
@@ -137,7 +147,7 @@ class AmazonService(OnlineService):
 
             self.save_user_data(user_data)
 
-            self.emit("service-login")
+            SERVICE_LOGIN.fire(self)
 
     def is_connected(self):
         """Return whether the user is authenticated and if the service is available"""
@@ -148,27 +158,17 @@ class AmazonService(OnlineService):
 
     def load(self):
         """Load the user game library from the Amazon API"""
-        if self.is_loading:
-            logger.warning("Amazon games are already loading")
-            return
         if not self.is_connected():
             logger.error("User not connected to Amazon")
             return
-        self.is_loading = True
-        try:
-            games = [AmazonGame.new_from_amazon_game(game) for game in self.get_library()]
-            for game in games:
-                game.save()
-        except:
-            logger.error("Unable to get games library")
-            games = None
-
-        self.is_loading = False
+        games = [AmazonGame.new_from_amazon_game(game) for game in self.get_library()]
+        for game in games:
+            game.save()
         return games
 
     def save_user_data(self, user_data):
         """Save the user data file"""
-        with open(self.user_path, "w", encoding='utf-8') as user_file:
+        with open(self.user_path, "w", encoding="utf-8") as user_file:
             user_file.write(json.dumps(user_data))
 
     def load_user_data(self):
@@ -176,9 +176,9 @@ class AmazonService(OnlineService):
         user_data = None
 
         if not os.path.exists(self.user_path):
-            raise AuthenticationError("No Amazon user data available, please log in again")
+            raise AuthenticationError(_("No Amazon user data available, please log in again"))
 
-        with open(self.user_path, "r", encoding='utf-8') as user_file:
+        with open(self.user_path, "r", encoding="utf-8") as user_file:
             user_data = json.load(user_file)
 
         return user_data
@@ -240,11 +240,11 @@ class AmazonService(OnlineService):
         try:
             request.post(json.dumps(data).encode())
         except HTTPError as ex:
-            logger.error("Failed http request %s", url)
-            raise AuthenticationError("Unable to register device, please log in again") from ex
+            logger.exception("Failed http request %s: %s", url, ex)
+            raise AuthenticationError(_("Unable to register device, please log in again")) from ex
 
         res_json = request.json
-        logger.info("Succesfully registered a device")
+        logger.info("Successfully registered a device")
         user_data = res_json["response"]["success"]
         return user_data
 
@@ -256,7 +256,7 @@ class AmazonService(OnlineService):
         expires_in = user_data["tokens"]["bearer"]["expires_in"]
 
         if not token_obtain_time or not expires_in:
-            raise AuthenticationError("Invalid token info found, please log in again")
+            raise AuthenticationError(_("Invalid token info found, please log in again"))
 
         return time.time() > token_obtain_time + int(expires_in)
 
@@ -289,8 +289,8 @@ class AmazonService(OnlineService):
         try:
             request.post(json.dumps(request_data).encode())
         except HTTPError as ex:
-            logger.error("Failed http request %s", url)
-            raise AuthenticationError("Unable to refresh token, please log in again") from ex
+            logger.exception("Failed http request %s: %s", url, ex)
+            raise AuthenticationError(_("Unable to refresh token, please log in again")) from ex
 
         res_json = request.json
 
@@ -330,9 +330,9 @@ class AmazonService(OnlineService):
 
         try:
             request.get()
-        except HTTPError:
+        except HTTPError as ex:
             # Do not raise exception here, should be managed from the caller
-            logger.error("Failed http request %s", url)
+            logger.exception("Failed http request %s: %s", url, ex)
             return False
 
         return True
@@ -341,7 +341,7 @@ class AmazonService(OnlineService):
         """Return the user's library of Amazon games"""
         if system.path_exists(self.cache_path):
             logger.debug("Returning cached Amazon library")
-            with open(self.cache_path, "r", encoding='utf-8') as amazon_cache:
+            with open(self.cache_path, "r", encoding="utf-8") as amazon_cache:
                 return json.load(amazon_cache)
 
         access_token = self.get_access_token()
@@ -349,41 +349,65 @@ class AmazonService(OnlineService):
         user_data = self.load_user_data()
         serial = user_data["extensions"]["device_info"]["device_serial_number"]
 
-        games = []
-        nextToken = None
+        games_by_asin = defaultdict(list)
+        next_token = None
         while True:
-            request_data = self.get_sync_request_data(serial, nextToken)
+            request_data = self.get_sync_request_data(serial, next_token)
 
-            json_data = self.request_sds(
-                "com.amazonaws.gearbox."
-                "softwaredistribution.service.model."
-                "SoftwareDistributionService.GetEntitlementsV2",
+            json_data = self.request_entitlements(
+                "com.amazon.animusdistributionservice.entitlement.AnimusEntitlementsService.GetEntitlements",
                 access_token,
                 request_data,
             )
 
-            if not json_data:
-                return
+            for game_json in json_data["entitlements"]:
+                product = game_json["product"]
 
-            games.extend(json_data["entitlements"])
+                if "id" not in product:
+                    logger.error("Amazon game encountered with no ID; skipping: %s", game_json)
+                    continue
+
+                # Some games have no ASIN; we'll have to skip dedupping these by ASIN.
+                asin = product.get("asin") or ""
+                games_by_asin[asin].append(game_json)
 
             if "nextToken" not in json_data:
                 break
 
             logger.info("Got next token in response, making next request")
-            nextToken = json_data["nextToken"]
+            next_token = json_data["nextToken"]
 
-        with open(self.cache_path, "w", encoding='utf-8') as amazon_cache:
+        # We need to use the ID for compatibility with earlier Lutris releases, but also
+        # some games do not have ASINs. We still deduplicate by ASIN when we can, since the same
+        # game can have two different IDs.
+
+        asinless = games_by_asin.pop("", [])
+        dedupped = [sorted(gl, key=lambda g: g["id"])[0] for gl in games_by_asin.values()]
+        games = asinless + dedupped
+
+        with open(self.cache_path, "w", encoding="utf-8") as amazon_cache:
             json.dump(games, amazon_cache)
 
         return games
 
-    def get_sync_request_data(self, serial, nextToken=None):
+    def request_distribution(self, target, token, body):
+        headers = {
+            "X-Amz-Target": target,
+            "x-amzn-token": token,
+            "UserAgent": "com.amazon.agslauncher.win/3.0.9202.1",
+            "Content-Type": "application/json",
+            "Content-Encoding": "amz-1.0",
+        }
+        request = Request(self.amazon_gaming_distribution, headers=headers)
+        request.post(json.dumps(body).encode())
+        return request.json
+
+    def get_sync_request_data(self, serial, next_token=None, sync_point=None):
         request_data = {
-            "Operation": "GetEntitlementsV2",
+            "Operation": "GetEntitlements",
             "clientId": "Sonic",
-            "syncPoint": None,
-            "nextToken": nextToken,
+            "syncPoint": sync_point,
+            "nextToken": next_token,
             "maxResults": 50,
             "productIdFilter": None,
             "keyId": "d5dc8b8b-86c8-4fc4-ae93-18c0def5314d",
@@ -397,21 +421,27 @@ class AmazonService(OnlineService):
             "X-Amz-Target": target,
             "x-amzn-token": token,
             "User-Agent": self.user_agent,
-            "UserAgent": self.user_agent,
             "Content-Type": "application/json",
             "Content-Encoding": "amz-1.0",
         }
 
         url = f"{self.amazon_sds}/amazon/"
         request = Request(url, headers=headers)
+        request.post(json.dumps(body).encode())
+        return request.json
 
-        try:
-            request.post(json.dumps(body).encode())
-        except HTTPError:
-            # Do not raise exception here, should be managed from the caller
-            logger.error("Failed http request %s", url)
-            return
+    def request_entitlements(self, target, token, body):
+        headers = {
+            "X-Amz-Target": target,
+            "x-amzn-token": token,
+            "User-Agent": self.user_agent,
+            "Content-Type": "application/json",
+            "Content-Encoding": "amz-1.0",
+        }
 
+        url = f"{self.amazon_gaming_entitlements}"
+        request = Request(url, headers=headers)
+        request.post(json.dumps(body).encode())
         return request.json
 
     def get_game_manifest_info(self, game_id):
@@ -419,24 +449,20 @@ class AmazonService(OnlineService):
         access_token = self.get_access_token()
 
         request_data = {
-            "adgGoodId": game_id,
-            "previousVersionId": None,
-            "keyId": "d5dc8b8b-86c8-4fc4-ae93-18c0def5314d",
-            "Operation": "GetDownloadManifestV3",
+            "entitlementId": game_id,
+            "Operation": "GetGameDownload",
         }
 
-        response = self.request_sds(
-            "com.amazonaws.gearbox."
-            "softwaredistribution.service.model."
-            "SoftwareDistributionService.GetDownloadManifestV3",
-            access_token,
-            request_data,
-        )
-
-        if not response:
-            logger.error("There was an error getting game manifest: %s", game_id)
-            raise UnavailableGame(
-                "Unable to get game manifest info, please check your Amazon credentials and internet connectivity")
+        try:
+            response = self.request_distribution(
+                "com.amazon.animusdistributionservice.external.AnimusDistributionService.GetGameDownload",
+                access_token,
+                request_data,
+            )
+        except HTTPError as ex:
+            # Do not raise exception here, should be managed from the caller
+            logger.exception("There was an error getting game '%s' manifest: %s", game_id, ex)
+            raise UnavailableGameError(_("Unable to get game manifest info")) from ex
 
         return response
 
@@ -444,68 +470,79 @@ class AmazonService(OnlineService):
         """Get a game manifest"""
         headers = {
             "User-Agent": self.user_agent,
-            "UserAgent": self.user_agent
         }
 
-        url = manifest_info["downloadUrls"][0]
+        url = manifest_info["downloadUrl"]
+        url = urllib.parse.urlparse(url)
+        url = url._replace(path=url.path + "/manifest.proto")
+        url = urllib.parse.urlunparse(url)
+
         request = Request(url, headers=headers)
 
         try:
             request.get()
         except HTTPError as ex:
-            logger.error("Failed http request %s", url)
-            raise UnavailableGame(
-                "Unable to get game manifest, please check your Amazon credentials and internet connectivity") from ex
+            logger.exception("Failed http request %s: %s", url, ex)
+            raise UnavailableGameError(_("Unable to get game manifest")) from ex
 
         content = request.content
 
         header_size = struct.unpack(">I", content[:4])[0]
 
         header = ManifestHeader()
-        header.decode(content[4: 4 + header_size])
+        header.decode(content[4 : 4 + header_size])
 
         if header.compression.algorithm == CompressionAlgorithm.none:
-            raw_manifest = content[4 + header_size:]
+            raw_manifest = content[4 + header_size :]
         elif header.compression.algorithm == CompressionAlgorithm.lzma:
-            raw_manifest = lzma.decompress(content[4 + header_size:])
+            raw_manifest = lzma.decompress(content[4 + header_size :])
         else:
             logger.error("Unknown compression algorithm found in manifest")
-            raise UnavailableGame(
-                "Unknown compression algorithm found in manifest, "
-                "please check your Amazon credentials and internet connectivity")
+            raise UnavailableGameError(_("Unknown compression algorithm found in manifest"))
 
         manifest = Manifest()
         manifest.decode(raw_manifest)
 
         return manifest
 
-    def get_game_patches(self, game_id, version, file_list):
-        """Get game files"""
-        access_token = self.get_access_token()
-
+    def get_file_patch(self, access_token, game_id, version, file_hashes):
         request_data = {
             "Operation": "GetPatches",
             "versionId": version,
-            "fileHashes": file_list,
+            "fileHashes": file_hashes,
             "deltaEncodings": ["FUEL_PATCH", "NONE"],
             "adgGoodId": game_id,
         }
 
-        response = self.request_sds(
-            "com.amazonaws.gearbox."
-            "softwaredistribution.service.model."
-            "SoftwareDistributionService.GetPatches",
-            access_token,
-            request_data,
-        )
+        try:
+            return self.request_sds(
+                "com.amazonaws.gearbox." "softwaredistribution.service.model." "SoftwareDistributionService.GetPatches",
+                access_token,
+                request_data,
+            )
+        except HTTPError as ex:
+            # Do not raise exception here, should be managed from the caller
+            logger.exception("There was an error getting '%s' patches: %s", game_id, ex)
+            raise UnavailableGameError(_("Unable to get the patches of game '%s'") % game_id) from ex
 
-        if not response:
-            logger.error("There was an error getting patches: %s", game_id)
-            raise UnavailableGame(
-                "Unable to get the patches of game, "
-                "please check your Amazon credentials and internet connectivity", game_id)
+    def get_game_patches(self, game_id, version, file_list):
+        """Get game files"""
+        access_token = self.get_access_token()
 
-        return response["patches"]
+        def get_batches(to_batch, batch_size):
+            i = 0
+            while i < len(to_batch):
+                yield to_batch[i : i + batch_size]
+                i += batch_size
+
+        batches = get_batches(file_list, 500)
+        patches = []
+
+        for batch in batches:
+            response = self.get_file_patch(access_token, game_id, version, batch)
+            patches += response["patches"]
+
+        return patches
 
     def structure_manifest_data(self, manifest):
         """Transform the manifest to more convenient data structures"""
@@ -520,11 +557,12 @@ class AmazonService(OnlineService):
                 hashes.append(file_hash)
                 files.append({"path": file.path.decode().replace("\\", "/"), "size": file.size, "url": None})
 
-                hashpairs.append(dict(
-                    sourceHash=None,
-                    targetHash=dict(value=file_hash,
-                                    algorithm=HashAlgorithm.get_name(file.hash.algorithm)),
-                ))
+                hashpairs.append(
+                    {
+                        "sourceHash": None,
+                        "targetHash": {"value": file_hash, "algorithm": HashAlgorithm.get_name(file.hash.algorithm)},
+                    }
+                )
             for __, directory in enumerate(package.dirs):
                 if directory.path is not None:
                     directories.append(directory.path.decode().replace("\\", "/"))
@@ -541,9 +579,12 @@ class AmazonService(OnlineService):
 
         file_dict, directories, hashpairs = self.structure_manifest_data(manifest)
 
-        game_patches = self.get_game_patches(game_id, manifest_info["versionId"], hashpairs)
-        for __, patch in enumerate(game_patches):
-            file_dict[patch["patchHash"]["value"]]["url"] = patch["downloadUrls"][0]
+        for file_hash, file in file_dict.items():
+            url = manifest_info["downloadUrl"]
+            url = urllib.parse.urlparse(url)
+            url = url._replace(path=url.path + "/files/" + file_hash)
+            url = urllib.parse.urlunparse(url)
+            file["url"] = url
 
         return file_dict, directories
 
@@ -551,7 +592,6 @@ class AmazonService(OnlineService):
         """Get and parse the fuel.json file"""
         headers = {
             "User-Agent": self.user_agent,
-            "UserAgent": self.user_agent
         }
 
         request = Request(fuel_url, headers=headers)
@@ -559,11 +599,19 @@ class AmazonService(OnlineService):
         try:
             request.get()
         except HTTPError as ex:
-            logger.error("Failed http request %s", fuel_url)
-            raise UnavailableGame(
-                "Unable to get fuel.json file, please check your Amazon credentials and internet connectivity") from ex
+            logger.error("Failed http request %s: %s", fuel_url, ex)
+            raise UnavailableGameError(_("Unable to get fuel.json file.")) from ex
 
-        res_json = request.json
+        try:
+            res_yaml_text = request.text
+            res_json = yaml.safe_load(res_yaml_text)
+        except Exception as ex:
+            # Maybe it can be parsed as plain JSON. May as well try it.
+            try:
+                logger.exception("Unparseable yaml response from %s: %s\n%s", fuel_url, ex, res_yaml_text)
+                res_json = json.loads(res_yaml_text)
+            except Exception:
+                raise UnavailableGameError(_("Invalid response from Amazon APIs")) from ex
 
         if res_json["Main"] is None or res_json["Main"]["Command"] is None:
             return None, None
@@ -586,48 +634,59 @@ class AmazonService(OnlineService):
             game_cmd, game_args = self.get_exe_and_game_args(fuel_url)
 
         if game_cmd is None:
-            game_cmd = "_xXx_AUTO_WIN32_xXx_"
+            game_cmd = AUTO_WIN32_EXE
 
         if game_args is None:
             game_args = ""
 
         return game_cmd, game_args
 
+    def get_installer_files(self, installer, _installer_file_id, _selected_extras):
+        try:
+            file_dict, __ = self.get_game_files(installer.service_appid)
+        except HTTPError as err:
+            raise UnavailableGameError(_("Couldn't load the downloads for this game")) from err
+
+        files = []
+        for file_hash, file in file_dict.items():
+            file_name = os.path.basename(file["path"])
+            files.append(
+                InstallerFile(
+                    installer.game_slug, file_hash, {"url": file["url"], "filename": file_name, "size": file["size"]}
+                )
+            )
+        # return should be a list of files, so we return a list containing a InstallerFileCollection
+        file_collection = InstallerFileCollection(installer.game_slug, "amazongame", files)
+        return [file_collection], []
+
+    def get_installed_slug(self, db_game):
+        details = json.loads(db_game["details"])
+        return slugify(details["product"]["title"])
+
     def generate_installer(self, db_game):
         """Generate a installer for the Amazon game"""
         details = json.loads(db_game["details"])
+        game_id = details["id"]
+
+        manifest_info = self.get_game_manifest_info(game_id)
+        manifest = self.get_game_manifest(manifest_info)
+
+        file_dict, directories, hashpairs = self.structure_manifest_data(manifest)
 
-        files = []
         installer = [
             {"task": {"name": "create_prefix"}},
-            {"mkdir": "$GAMEDIR/drive_c/game"}]
-
-        file_dict, directories = self.get_game_files(details["id"])
-
-        for __, directory in enumerate(directories):
-            installer.append({"mkdir": f"$GAMEDIR/drive_c/game/{directory}"})
+            {"mkdir": "$GAMEDIR/drive_c/game"},
+            {"autosetup_amazon": {"files": file_dict, "directories": directories}},
+        ]
 
+        # try to get fuel file that contain the main exe
+        fuel_file = [k for k, v in file_dict.items() if "fuel.json" in v["path"]]
         fuel_url = None
-
-        for file_hash, file in file_dict.items():
-            file_name = os.path.basename(file["path"])
-            files.append({
-                file_hash: {
-                    "url": file["url"],
-                    "filename": file_name,
-                    "provider": "download"
-                }
-            })
-
-            file_path = os.path.dirname(file["path"])
-            installer.append({"move": {
-                "description": _("Installing file: %s") % file_name,
-                "src": file_hash,
-                "dst": f"$GAMEDIR/drive_c/game/{file_path}"
-            }})
-
-            if file_name == "fuel.json":
-                fuel_url = file["url"]
+        if fuel_file:
+            fuel_url = manifest_info["downloadUrl"]
+            fuel_url = urllib.parse.urlparse(fuel_url)
+            fuel_url = fuel_url._replace(path=fuel_url.path + "/files/" + list(fuel_file)[0])
+            fuel_url = urllib.parse.urlunparse(fuel_url)
 
         game_cmd, game_args = self.get_game_cmd_line(fuel_url)
         logger.info("game cmd line: %s %s", game_cmd, game_args)
@@ -637,16 +696,19 @@ class AmazonService(OnlineService):
             "version": _("Amazon Prime Gaming"),
             "slug": slugify(details["product"]["title"]),
             "game_slug": slugify(details["product"]["title"]),
-            "runner": "wine",
+            "runner": self.get_installed_runner_name(db_game),
             "script": {
                 "game": {
                     "exe": f"$GAMEDIR/drive_c/game/{game_cmd}",
                     "args": game_args,
                     "prefix": "$GAMEDIR",
-                    "working_dir": "$GAMEDIR/drive_c/game"
+                    "working_dir": "$GAMEDIR/drive_c/game",
                 },
                 "system": {},
-                "files": files,
-                "installer": installer
-            }
+                "files": [{"amazongame": "N/A:Select the installer from Amazon Games"}],
+                "installer": installer,
+            },
         }
+
+    def get_installed_runner_name(self, db_game):
+        return self.runner
